package urpmchecker;

use strict;

our $VERSION="0.0.1";
my $CVS='$Id: urpmchecker.pm,v 1.13 2003/04/25 00:32:57 nanardon Exp $';

use vars qw($VERSION $CVS @ISA @EXPORT);

use URPM;
use URPM::Build;
use MDK::Common;
use BerkeleyDB;

@ISA = qw(VERSION Exporter URPM);

=head1 NAME

urpmchecker - function to works on rpms repository

=head1 SYMOPSYS

use urpmchecker
$urpmchecker = new urpmchecker

=head1 DESCRIPTION

C<urpmchecker> include urpmchecker functions
	
=head1 SEE ALSO

perl-URPM

=head1 LICENSE

Olivier Thauvin <thauvin@aerov.jussieu.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

=item * $urpmchecker->new($param);

$param = {
    verbose => 1, # set verbose flag
	rpmrc => '/usr/lib/rpm/rpmrc' #rpmrc to use
}
	
=cut

sub new {
	my ($self,$param)=@_;
	my $env = new BerkeleyDB::Env
		-Home => $ENV{TMPDIR} || $ENV{TMP} || "/tmp";
	#@{$param->{qw(dbfilesf dbprovidesf)}}=("/tmp/distriblint$$.files.db","/tmp/distriblint$$.provides.db");
	$param->{dbfiles} =  new BerkeleyDB::Btree
			#-Filename => $param->{dbfilesf},
                	-Flags    => DB_CREATE,
		            -Property => DB_DUP | DB_DUPSORT,
			    -Env => $env,
        		or die "Cannot open file $param->{dbfilesf}\: $! $BerkeleyDB::Error\n";
	$param->{dbprovides} = new BerkeleyDB::Btree
			#-Filename => $param->{dbprovidesf},
                	-Flags    => DB_CREATE,
		            -Property => DB_DUP | DB_DUPSORT,
			    -Env => $env,
        		or die "Cannot open file $param->{dbprovidesf}\: $! $BerkeleyDB::Error\n";
	$param->{rpmvalue}=parse_rc($param->{rpmrc} ? $param->{rpmrc} : "/usr/lib/rpm/rpmrc");
	
	bless {
		urpm => URPM->new(),
		dbfiles => $param->{dbfiles},
		dbfilesf => $param->{dbfilesf},
		dbprovides => $param->{dbprovides},
		dbprovidesf => $param->{dbprovidesf},
		dbobsoletes => {},
		obsoletes => {},
		deps => {},
		repository => [],
		last_rep_id => 0,
		id_rep => {},
		error => {},
		verbose => $param->{verbose},
		rpmvalues => $param->{rpmvalue},
	}, $self;
}

sub DESTROY {
   my ($self)=@_;
   $self->{dbfiles}->db_close();
   $self->{dbprovides}->db_close();
   unlink($self->{dbfilesf},$self->{dbprovidesf}); 
}

=item * $urpmchecker::version();

return module version.

=cut 

sub version {
	"Version: $VERSION; $CVS\n"
}

sub add {
	my ($self,$item,$param)=@_;
	$_=$item;
	$param->{nextid} and $self->{last_rep_id}++;
	print STDERR "Add $_ in list\n" if ($self->{verbose});
	my $to_add = { 
			check => $param->{check}, 
			id => $self->{last_rep_id}, 
			origin => $_,
			checksrc => $param->{checksrc},
			chkdep => $param->{checkdep},
		    chkcf => $param->{checkcf},
			};
			
	if (/cz$/) { 
	   -f $_ or return(undef);
	   	  $to_add->{hdlist} = $_;
        } elsif (/\.rpm$/) {
	   -f $_ or return(undef);
           @{$to_add->{rpms}} = ( $_ );
       } else {
	   	  my @l=glob "$_/*.rpm";
	   	  scalar(@l) or return(undef);
       	  @{$to_add->{rpms}} = @l;
   }
   push @{$self->{repository}},$to_add;
   1
}

=item * $urpmchecker->parselist();

For each repository added, parse rpm headers and run URPM::compute_dep.

=cut

sub parse_list {
	my ($self)=@_;
	foreach(@{$self->{repository}}) {
		if ($_->{hdlist}) {
		($_->{start}, $_->{end}) = $self->{urpm}->parse_hdlist($_->{hdlist});
		push @{$_->{list}}, $_->{start} .. $_->{end};
    		}
		@{$_->{list}} = map { $self->{urpm}->parse_rpm($_, keep_all_tags => 1) } @{$_->{rpms}} if $_->{rpms};
	}
	
	foreach my $l (@{$self->{repository}}) {
	@{$l->{list}}=uniq(@{$l->{list}});
	#removing non compatible rpm
	@{$l->{list}}= grep { 
		my $pkg = $self->{urpm}->{depslist}[$_]; 
		$self->rpmcompat($pkg->arch,"arch_compat") and $_
		} @{$l->{list}};
	$self->{id_rep}->{$_}=$l->{id} foreach(@{$l->{list}});
	}
	
	foreach my $l (@{$self->{repository}}) {
		foreach my $id (@{$l->{list}}) {
			my $pkg = $self->{urpm}->{depslist}[$id];
			my ($name,$version,$release,$arch);
			if ($pkg->arch eq 'src') {
				($name,$version,$release)=($pkg->name,$pkg->version,$pkg->release);
				$arch = 'src';
			} else {
				my $sourcerpm=$pkg->sourcerpm;
				($name,$version,$release)= $sourcerpm =~ /^(.*)-([^\-]*)-(.*)\.src.rpm$/;
				$arch ='bin';
			}
			# print $pkg->name." $arch $name $version-$release $id\n";
			push(@{$self->{rpm}->{$arch}->{$name}->{"$version-$release"}},$id);
			
			my @o = $pkg->obsoletes;
			foreach (@o) {
				my ($n, $o, $v) = /^([^\s\[]*)(?:\[\*\])?\s*\[?([^\s\]]*)\s*([^\s\]]*)/;
				push(@{$self->{dbobsoletes}->{$n}},"$id $o $v");
			}
			# All info after does concern src... why wast time
			# is a src can obsoletes another ??? That is the question. It seems no
			$pkg->arch eq 'src' and next;
			my @f = $pkg->files;
			my @m = $pkg->files_mode;
			my @u = $pkg->files_owner;
			my @g = $pkg->files_group;
			my @md5 = $pkg->files_md5sum;
			foreach (0 .. $#f) {
				$self->{dbfiles}->db_put($f[$_],"$id $m[$_] $md5[$_] $u[$_] $g[$_]");
			}
			my @p = $pkg->provides;
			foreach (@p) {
				my ($n, $o, $v) = /^([^\s\[]*)(?:\[\*\])?\s*\[?([^\s\]]*)\s*([^\s\]]*)/;
				$self->{dbprovides}->db_put($n,"$id $o $v");
			}
		}
	}
	foreach my $l (@{$self->{repository}}) {
		foreach my $id (@{$l->{list}}) {
			my $pkg=$self->{urpm}->{depslist}[$id];
			#$self->rpmcompat($pkg->arch,"arch_compat") or next;
			# Parsing dependencies (Requires <=> Provides/Files)
			foreach ($pkg->requires) {
				my ($n, $o, $v) = /^([^\s\[]*)(?:\[\*\])?\s*\[?([^\s\]]*)\s*([^\s\]]*)/;
				$o =~ s/==/=/g;
				my ($value,$cursor_state,@tdeps)=(undef,DB_SET,());
				if ($n =~ /^\//) {
					my $cursor = $self->{dbfiles}->db_cursor();
					while ($cursor->c_get($n,$value,$cursor_state)==0) {
						push(@tdeps,(split(/ /,$value))[0]);
						$cursor_state=DB_NEXT_DUP;	
					}			
		    	} else {
					my $cursor = $self->{dbprovides}->db_cursor();
					while ($cursor->c_get($n,$value,$cursor_state)==0) {
						my ($p, $po, $pv)=split(' ', $value);
						$po =~ s/==/=/g;
						# It want a name, well, we not loose time with long name, we allready make it ;)
						URPM::ranges_overlap("a $po $pv","a $o $v") and push(@tdeps,$p);
						$cursor_state=DB_NEXT_DUP;	
					}
				}
			(scalar @tdeps) ? push(@{$self->{deps}->{$id}},join('|',@tdeps)) : push(@{$self->{unresolved}->{$id}},$_);
			}
			# Parsing obsoletes
			foreach (@{$self->{dbobsoletes}->{$pkg->name}}) {
				my ($p, $o, $v)=split(' ');
				$o && ! eval($pkg->compare($v) . $o . 0) and next;
				($pkg->arch eq 'src' || $self->{urpm}->{depslist}[$p]->arch eq 'src') and
					($pkg->arch ne $self->{urpm}->{depslist}[$p]->arch)	and next;
				push(@{$self->{obsoletes}->{$id}},$p);
			}
			@{$self->{obsoletes}->{$id}}=uniq(@{$self->{obsoletes}->{$id}});
		}
	}
}
sub analyze {
	my ($self,$options)=@_;
	$self->analyze_dep();
	$self->analyze_files();
	$self->analyze_src();
}

sub analyze_dep {
	my ($self)=@_;
	my $cursor;
	foreach my $l (@{$self->{repository}}) {
		$l->{check} or next;
		$l->{chkdep} and next;
		foreach my $id (@{$l->{list}}) {
			my $pkg=$self->{urpm}->{depslist}[$id];
			#$self->rpmcompat($pkg->arch,"arch_compat") or next;
			foreach (@{$self->{unresolved}->{$id}}) {
				my ($n, $o, $v) = /^([^\s\[]*)(?:\[\*\])?\s*\[?([^\s\]]*)\s*([^\s\]]*)/;
				$self->report_error($id,{ msg => "$n $o $v", type => 'DEP ' }); 
			}
			foreach (@{$self->{deps}->{$id}}) {
				my $r=min(map { $self->{id_rep}->{$_} } split(/\|/));
				$self->report_error($id,{ id => $_, type => 'DEPO', msg => "req in $r" }) if ($l->{id} < $r);
			}
			foreach(@{$self->{obsoletes}->{$id}}) {
				$self->report_error($id,{ id => $_, type => 'OBS ', msg => "obs in $self->{id_rep}->{$_}" }) if ($l->{id} >= $self->{id_rep}->{$_});
			}
			$pkg->arch eq 'src' and next;
			my ($name,$version,$release)= $pkg->sourcerpm =~ /^(.*)-([^\-]*)-([^\.]*)\.src.rpm$/;
			scalar keys %{$self->{rpm}->{bin}->{$name}} > 1 and do {
				foreach my $vr (keys %{$self->{rpm}->{bin}->{$name}}) {
					$vr eq "$version-$release" and next;
					#my $idem;
					my @idem = grep { 
						my $pkg2=$self->{urpm}->{depslist}[$_];
						$pkg->name eq $pkg2->name and $_
						} @{$self->{rpm}->{bin}->{$name}->{$vr}};
					if (scalar @idem) {
						foreach (@idem) {
							$_ == $id and next;
							my $pkg2=$self->{urpm}->{depslist}[$_];
							$pkg->name eq $pkg2->name or next;
							$pkg->compare_pkg($pkg2) > 0 or $self->report_error($id,{ id => $_, type => 'VER ', msg => "new ver in $self->{id_rep}->{$_}" });
						# print $pkg->fullname." ".$pkg2->fullname." ".$pkg->compare_pkg($pkg2)." @idem\n";
						}
					} else {
						#don't find exact same name, but there are a pb with version checking
					}
				};
			}
		}
	}
}	

sub analyze_files {
	my ($self)=@_;	
	#my (%files)=(0);
	
	my ($cursor,$file,$filekey) = $self->{dbfiles}->db_cursor();
	while ($cursor->c_get($file, $filekey, DB_NEXT) == 0) {
		my @l;
		$filekey =~ m/(\S*) (\S*) (\S*) (\S*) (\S*)/; 
		push(@l, { id => $1, mode => $2, md5sum => $3, user => $4, group => $5 });
		while ($cursor->c_get($file, $filekey, DB_NEXT_DUP) == 0) {
			$filekey =~ m/(\S*) (\S*) (\S*) (\S*) (\S*)/; 
			push(@l, { id => $1, mode => $2, md5sum => $3, user => $4, group => $5 });	
		}
		
		foreach my $id (@l) {
			
			# Is the parent is provide by another rpm
			my ($reper,$refile)=$file =~  m/^(.*)\/(.*)/;
			my $cursor2=$self->{dbfiles}->db_cursor();
						
			my ($cursor_state,$value,@d)=(DB_SET);
			while ($cursor2->c_get($reper,$value,$cursor_state)==0) {
				my ($p,$mode)= $value =~ m/(\S*) (\S*) (\S*) (\S*) (\S*)/;
				($mode & 0170000) == 0040000 and push(@d,$p);
				$cursor_state=DB_NEXT_DUP;	
			}
			if (@d) {
				my $min=min(map { $self->{id_rep}->{$_} }@d);
				$min > $self->{id_rep}->{$id->{id}} and $self->report_error($id->{id},{ type => 'DIRO', msg => "file $refile, rep in $min", msgsum => "$reper/" });
			} else {
				$self->report_error($id->{id},{ type => 'DIRM', msg => "file $refile", msgsum => "$reper/" });
			}
			foreach (@l) {
				# Check rpm against rpm
				$_->{id} eq $id->{id} and next; # Check itself, no !
					
				# same name, surelly an update
				$self->{urpm}->{depslist}[$_->{id}]->name eq $self->{urpm}->{depslist}[$id->{id}]->name and next;

				# Is those packages conflicts ?
				pkg_is_conflict($self->{urpm}->{depslist}[$_->{id}],$self->{urpm}->{depslist}[$id->{id}]) and next;	
				# Now checking files against files, here the tags:	
				#define __S_IFDIR       0040000 /* Directory.  */
				#define __S_IFCHR       0020000 /* Character device.  */
				#define __S_IFBLK       0060000 /* Block device.  */
				#define __S_IFREG       0100000 /* Regular file.  */
				#define __S_IFIFO       0010000 /* FIFO.  */
				#define __S_IFLNK       0120000 /* Symbolic link.  */
				#define __S_IFSOCK      0140000 /* Socket.  */
						
				($_->{mode} & 0170000) != ($id->{mode} & 0170000) and do { 
					$self->report_error($id->{id},{ id => $_->{id}, type => 'CFLT', msg => "$file" }); 
					next 
					};
				($_->{mode} & 0006666) != ($id->{mode} & 0006666) and
					$self->report_error($id->{id},{ id => $_->{id}, type => 'CFLP', msg => "$file" });
				($_->{user} ne $id->{user} or $_->{group} ne $id->{group}) and
					$self->report_error($id->{id},{ id => $_->{id}, type => 'CFLO', msg => "$file" });
				($id->{mode} & 0170000) == 0040000 and next;
						
				# same md5sum, next
				$_->{md5sum} == $id->{md5sum} and next;
								
				$self->report_error($id->{id},{ id => $_->{id}, type => 'CFL ', msg => "$file" });
			}
		}
	}
	undef $cursor;
}

sub analyze_src {
	my ($self)=@_;
	foreach my $repository (@{$self->{repository}}) {
		$repository->{check} or next;
		$repository->{checksrc} or next;
		foreach my $id (@{$repository->{list}}) {
			my $pkg=$self->{urpm}->{depslist}[$id];
			if ($pkg->arch eq 'src') {
				exists $self->{rpm}->{bin}->{$pkg->name} or $self->report_error($id, { type => 'BIN ', msg => "no binary from ". $pkg->name });
				foreach(@{$self->{rpm}->{bin}->{$pkg->name}->{$pkg->version.'-'.$pkg->release}}) {
					$self->{id_rep}->{$_} == $repository->{id} or 
						$self->report_error($id, { type => 'BINO', msg => sprintf("bin in %s",$self->{id_rep}->{$_}), id => $_ });
				}
				
			} else {
				my ($name,$version,$release)= $pkg->sourcerpm =~ /^(.*)-([^\-]*)-([^\.]*)\.src.rpm$/;
				#print "$name - $version - $release\n";
				$self->{rpm}->{src}->{$name} or $self->report_error($id, { type => 'SRC ', msg => "no src named $name" });
				foreach(@{$self->{rpm}->{src}->{$name}->{"$version-$release"}}) {
					$self->{id_rep}->{$_} == $repository->{id} or 
						$self->report_error($id, { type => 'SRCO', msg => "src in $self->{id_rep}->{$_}", id => $_ });
				}

			}
		}
	}
	
}

=item * $urpmchecker->report_error($self,$id,$error);
$id = id of rpm
error = {
	id = id of rpm
	fullname = name of rpm
	+packager => from latest changelog
	+sourcerpm => the src
	@err = 
		{ 
		id => rpm wich conflict
		+fullname => of package wich conflict
		type => kind of error
		msgsum => short msg error
		msg => full error
}

=cut

sub report_error {
	my ($self,$id,$error)=@_;
	if (exists $error->{id} && exists $self->{urpm}->{depslist}[$error->{id}]) {
		$error->{fullname}=$self->{urpm}->{depslist}[$error->{id}]->fullname;
	}
	push(@{$self->{error}->{$id}->{err}},$error);
	
	if (!exists $self->{error}->{$id}->{fullname}) {
			$self->{error}->{$id}->{fullname}=$self->{urpm}->{depslist}[$id]->fullname;
			$self->{urpm}->{depslist}[$id]->arch ne 'src' and $self->{error}->{$id}->{sourcerpm}=$self->{urpm}->{depslist}[$id]->sourcerpm;
			my @c = $self->{urpm}->{depslist}[$id]->changelog_name;
			$c[0] =~ /^(.*) /;
			$self->{error}->{$id}->{packager}=$1;
			$self->{error}->{$id}->{level}=$self->{id_rep}->{$id};
	}
}

=item * urpmchecker::explanation_error(@ERR);

Return an array of explanation for error index in @ERR.
Without @ERR, return know ERR.

=cut

sub explanation_error {
	my %err = (
DIRO =>
"A file is installed but the rpm which parent directory is in
repository level highter. To cleanly install, you should this level",
DIRM =>
"A file is installed by a rpm but no rpm own parent dir.
This mean a uninstall the rpm, the parent dir will be removed",
CFL =>
"Those rpms contains same file, but files are differents, and those rpm do
not explicitly conflict. rpm will be unable to handle this safety",
CFLT => 
"Those rpms contains same file, but files have differents type
(dir vs link, file vs link), those rpm do not explicitly conflict,
rpm will be unable to handle this safety",
CFLP =>
"Those rpms contains same files but files but without same permissions.
If those files come from server apps, or contains secret passwd, 
it can create a security hole",
CFLO =>
"Those rpms contains same files but files are not own by same user or group.
If those files come from server apps, or contains secret passwd,
it can create a security hole",
DEP =>
"This requirement is missing in the distrib, installation or
building is impossible, maybe it need a rebuild",
DEPO =>
"This requirement is present but in a sublevel repository, you can
install or build it only if you have the second repository in your config",
OBS =>
"The package in level x is always obsolete by another in level y, 
then why it is in distro? If a package obsolete itself, there is a problem :)",
BIN =>
"This src package does produce any binary...",
BINO =>
"This src package produce binary but in another repository",
SRC =>
"This binary does not have a src",
SRC0 =>
"This binary have a src but in another repository",
VER =>
"An highter version of this package was found, so it shoulb be removed",
	);
	@_ or return "Know error: ".join(" ",sort keys %err)."\n";
	map { "$_\n" . $err{$_}."\n" } @_
}

sub parse_rc {
	my ($file)=@_;
	my $rpmvalue;
	open RC, $file or return 0;
	while (<RC>) {
		s/#.*//;         # no comments
        s/^\s+//;        # no leading white
        s/\s+$//;        # no trailing white
		length or next;
		my ($tag,$rest)= m/^([^:]*):(.*)$/;
		$tag =~ /^(arch_canon)$/ and do {
			my ($key,$compat,$val)= $rest =~ m/^\s*([^:]*):\s*(\S*)\s*(\S*)/;
			$rpmvalue->{$tag}->{$key}->{$compat}=$val ? $val : 1;
			next;
			};
		$tag =~ /^(arch_compat)|(buildarch_compat)$/ and do {
			my ($key,$rest)= $rest =~ m/^\s*([^:]*):\s*(.*)\s*$/;
			$rpmvalue->{$tag}->{$key}=$_ foreach(split(/\s+/,$rest));
			next;
			};
	}	
	close RC;
	$rpmvalue
}

sub rpmcompat {
	my ($self,$rpm_arch,$type,$os_arch)=@_;
	my $level=1;
	$os_arch or $os_arch=$self->{arch};
	# this function is for binary, and we want the reference system !
	$os_arch or return 1;
	$rpm_arch =~ /src$/ and return 1;
	$rpm_arch eq $os_arch and return $level;
	while (exists $self->{rpmvalues}->{$type}->{$os_arch}) {
		$level++;
		$self->{rpmvalues}->{$type}->{$os_arch} eq $rpm_arch and return $level;
		$os_arch=$self->{rpmvalues}->{$type}->{$os_arch};
	}
	0
}

sub pkg_is_conflict {
    my ($un,$deux)=@_;
    foreach my $c ($un->conflicts) {
        foreach($deux->provides) {
            URPM::ranges_overlap($c,$_) and return 1;
        }
    }
    foreach my $c ($deux->conflicts) {
        foreach ($un->provides) {
            URPM::ranges_overlap($c,$_) and return 1;
        }
    }
0
}

1
